from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, List, Optional, Self, Type, override

from pyparsing import ParseException, nestedExpr

import cqp_tree.translation as ct


def parse(s: str):
    # Parsing adapted from:
    # https://github.com/aarneranta/deptreepy/blob/a3fd7aa0b01f169afe6f37277d8bc2c624bcb433/patterns.py#L334
    if not s.startswith('('):  # add outer parentheses if missing
        s = '(' + s + ')'
    try:
        parsed = nestedExpr().parseString(s)
    except ParseException as ex:
        error = ct.InputError(f'line: {ex.line}, col: {ex.col}', ex.msg)
        raise ct.ParsingFailed(error)

    def to_lisp(lisp):
        match lisp:
            case [*args]:
                return [to_lisp(arg) for arg in args]
            case tok:
                return tok

    return to_lisp(parsed[0])


class Result(ABC):

    @abstractmethod
    def as_query(self, builder: ct.Recipe.Builder) -> 'Query': ...

    @abstractmethod
    def as_dependency_constraint(self) -> 'DependencyConstraint': ...


@dataclass
class Query(Result):
    identifier: ct.Identifier

    @override
    def as_query(self, builder: ct.Recipe.Builder) -> Self:
        return self

    @override
    def as_dependency_constraint(self) -> 'DependencyConstraint':
        raise ct.NotSupported('This operation does not support ')


class DependencyConstraint(Result):
    root: ct.Token

    tokens: List[ct.Token]
    dependencies: List[ct.Dependency]

    def __init__(self, token: ct.Token):
        self.root = token
        self.tokens = [token]
        self.dependencies = []

    def add_edge_to(self, target: Result):
        constraint = target.as_dependency_constraint()

        dependency = ct.Dependency(
            src=self.root.identifier,
            dst=constraint.root.identifier,
        )
        self.dependencies.append(dependency)

        self.tokens += constraint.tokens
        self.dependencies += constraint.dependencies

    @override
    def as_query(self, builder: ct.Recipe.Builder) -> Query:
        query = ct.Query(tokens=self.tokens, dependencies=self.dependencies)
        return Query(builder.add_query(query))

    @override
    def as_dependency_constraint(self) -> Self:
        return self


@dataclass
class TokenConstraint(Result):
    predicate: Optional[ct.Predicate]

    @override
    def as_query(self, builder: ct.Recipe.Builder) -> Query:
        return self.as_dependency_constraint().as_query(builder)

    @override
    def as_dependency_constraint(self) -> DependencyConstraint:
        token = ct.Token(attributes=self.predicate)
        return DependencyConstraint(token)


def operation_constructor_for_field(field) -> Callable[[str], ct.Comparison]:
    if not isinstance(field, str):
        raise ct.NotSupported('When matching a field, the field must be a string.')

    comparison_operator = '='
    if field.endswith('_'):
        field = field[:-1]
        comparison_operator = 'contains'

    def constructor(strpatt) -> ct.Comparison:
        if not isinstance(strpatt, str):
            raise ct.NotSupported('When matching a field, the field value must be a string.')
        return ct.Comparison(
            ct.Attribute(None, field),
            comparison_operator,
            ct.Literal(f'"{strpatt}"', represents_regex=True),
        )

    return constructor


@ct.translator('deptreepy')
def translate_deptreepy(deptreepy: str) -> ct.Recipe:
    builder = ct.Recipe.Builder()

    def combine_operation(
        parts: List[Query | DependencyConstraint | TokenConstraint],
        ctor: Type[ct.Conjunction | ct.Disjunction],
        operator: ct.SetOperator,
    ) -> TokenConstraint | Query:
        if all(isinstance(part, TokenConstraint) for part in parts):
            conjuncts = [part.predicate for part in parts if part.predicate]
            pred = ctor(conjuncts)
            return TokenConstraint(predicate=pred)

        else:
            # Promote to queries
            parts = [part.as_query(builder).identifier for part in parts]

            res, *others = parts
            for conj in others:
                res = builder.add_operation(res, operator, conj)
            return Query(res)

    # we want to return here for every possible operator
    # pylint: disable=too-many-return-statements
    def convert(lisp) -> TokenConstraint | DependencyConstraint | Query:
        match lisp:
            case ['TREE', *_]:
                raise ct.NotSupported('Only TREE_ is supported for matching subtrees.')

            case [singleton]:
                return convert(singleton)

            case ['TREE_', *args]:
                if not args:
                    raise ct.NotSupported('TREE_ requires at least 1 argument, got 0.')
                root, *dependents = args

                # TODO: Does this handle nested TREE_ correctly?
                constraint = convert(root).as_dependency_constraint()
                for dep in dependents:
                    constraint.add_edge_to(convert(dep))

                return constraint

            case ['AND', *args]:
                conjuncts = [convert(arg) for arg in args]
                if not conjuncts:  # AND without any predicates will match every token.
                    return TokenConstraint(predicate=None)

                return combine_operation(conjuncts, ct.Conjunction, ct.SetOperator.CONJUNCTION)

            case ['OR', *args]:
                disjuncts = [convert(arg) for arg in args]
                if not disjuncts:
                    raise ct.NotSupported('Empty OR matches no token. This is not supported.')

                return combine_operation(disjuncts, ct.Disjunction, ct.SetOperator.DISJUNCTION)

            case ['NOT', *args]:
                arg, *more = args
                if more:
                    raise ct.NotSupported(f'NOT requires exactly 1 argument, got {len(more) + 1}.')

                negated = convert(arg)
                if not isinstance(negated, TokenConstraint):
                    # TODO: The best way to handle this might be a flag for negated Query instances.
                    # AND containing NOT can then be implemented as subtraction.
                    # NOT in OR / as root will be expensive, querying entire corpus and subtracting.
                    raise ct.NotSupported(
                        'Searching for the absence of dependencies using NOT is not yet supported.'
                    )

                predicate = ct.Negation(negated.predicate)
                return TokenConstraint(predicate=predicate)

            case [field, 'IN', *strpatts]:
                ctor = operation_constructor_for_field(field)
                pred = ct.Disjunction([ctor(strpatt) for strpatt in strpatts])
                return TokenConstraint(predicate=pred)

            case [field, strpatt]:
                pred = operation_constructor_for_field(field)(strpatt)
                return TokenConstraint(predicate=pred)

            case unsupported:
                raise ct.NotSupported(f'Encountered unsupported expression: {unsupported}')

    result = convert(parse(deptreepy))
    if isinstance(result, TokenConstraint):
        result = result.as_dependency_constraint()
    if isinstance(result, DependencyConstraint):
        result = result.as_query(builder)

    builder.set_goal(result.identifier)
    return builder.build()



import math
import unittest

import cqp_tree.translation.cqp as cqp
import cqp_tree.translation.query as query


class TranslatorTests(unittest.TestCase):

    def test_arrangements_without_constraints(self):
        identifiers = {query.Identifier() for _ in range(7)}

        arrangements = cqp.arrangements(identifiers, [])

        self.assertEqual(
            len(list(arrangements)),
            math.factorial(7),
            '7 independent tokens should be arrangeable in 7! different ways.',
        )

    def test_arrangements(self):
        a = query.Identifier()
        b = query.Identifier()
        c = query.Identifier()
        d = query.Identifier()

        possible_arrangements = [[a, b, c, d], [a, b, d, c], [a, d, b, c]]
        constraints = {
            query.Constraint.order(a, b),
            query.Constraint.order(b, c),
            query.Constraint.order(a, d),
        }

        arrangements = list(cqp.arrangements({a, b, c, d}, constraints))
        self.assertEqual(
            len(arrangements),
            len(possible_arrangements),
            f'There are {len(possible_arrangements)} possible arrangements.',
        )
        for arrangement in arrangements:
            self.assertIn(arrangement, possible_arrangements)



import itertools
from typing import Annotated, Callable, Collection, Iterable, Iterator, Tuple

type NonEmpty[T] = Annotated[Collection[T], 'Non empty Sequence of T.']


def flatmap_set[X, Y](xs: Iterable[X], f: Callable[[X], Iterable[Y]]) -> set[Y]:
    result = set()
    for x in xs:
        result.update(f(x))
    return result


def partition_set[X](s: set[X], predicate: Callable[[X], bool]) -> Tuple[set[X], set[X]]:
    """
    Partitions a set into two sets using the given predicate.
    """
    t, f = set(), set()
    for e in s:
        if predicate(e):
            t.add(e)
        else:
            f.add(e)
    return t, f


def to_str(strings: Iterable[str], prefix: str = '', sep: str = '', suffix: str = '') -> str:
    """
    Creates a string from all the elements separated using separator and
    using the given prefix and postfix if supplied.
    """
    return prefix + sep.join(strings) + suffix


def format_human_readable(strings: NonEmpty[str]) -> str:
    """
    Returns a human-readable representation of a list of strings,
    separating the first values by a comma and the last one with "and".
    """
    if len(strings) == 1:
        return next(iter(strings))
    else:
        *init, last = strings
        initial = ', '.join(init)
        return f'{initial} and {last}'


LOWERCASE_ALPHABET = 'abcdefghijklmnopqrstuvwxyz'
UPPERCASE_ALPHABET = LOWERCASE_ALPHABET.upper()


def names_from_alphabet(alphabet: Iterable[str]) -> Iterator[str]:
    """Function producing an infinite stream of fresh names."""
    assert alphabet, 'alphabet cannot be empty.'

    length = 0
    while True:
        length += 1
        for name in itertools.combinations_with_replacement(alphabet, length):
            yield ''.join(name)


def associate_with_names[X](xs: Collection[X], alphabet: Iterable[str]) -> dict[X, str]:
    return dict(zip(xs, names_from_alphabet(alphabet)))



import unittest

from cqp_tree.utils import *


class UtilsTest(unittest.TestCase):

    def test_names_generates_fresh_names(self):
        limit = 1000

        seen = set()
        for index, name in enumerate(names_from_alphabet('abc')):
            self.assertNotIn(name, seen, f'{name} should be a fresh name.')
            seen.add(name)

            if index == limit:
                break




from abc import ABC, abstractmethod
from argparse import Namespace
from dataclasses import dataclass, field
from typing import Iterable, Iterator, Optional

from cqp_tree.translation import query
from cqp_tree.utils import (
    LOWERCASE_ALPHABET,
    UPPERCASE_ALPHABET,
    associate_with_names,
    flatmap_set,
    partition_set,
    to_str,
)

Environment = dict[query.Identifier, str]

TOKEN_ALPHABET = LOWERCASE_ALPHABET
QUERY_ALPHABET = UPPERCASE_ALPHABET

CQP_OPERATIONS = {
    query.SetOperator.CONJUNCTION: 'intersect',
    query.SetOperator.DISJUNCTION: 'union',
    query.SetOperator.SUBTRACTION: 'diff',
}


class Query(ABC):
    """Abstract base class for all queries."""

    target_identifier: Optional[query.Identifier]

    @abstractmethod
    def referenced_identifiers(self) -> set[query.Identifier]: ...

    @abstractmethod
    def format(self, environment: Environment) -> str: ...

    def __str__(self):
        environment = associate_with_names(self.referenced_identifiers(), TOKEN_ALPHABET)
        return self.format(environment)


@dataclass
class Sequence(Query):
    """Class representing a sequence of queries."""

    lhs: Query
    rhs: Query

    def referenced_identifiers(self) -> set[query.Identifier]:
        return self.lhs.referenced_identifiers() | self.rhs.referenced_identifiers()

    def format(self, environment: Environment) -> str:
        def format_subquery(q: Query) -> str:
            res = q.format(environment)
            return f'({res})' if isinstance(q, Operator) else res

        lhs_repr = format_subquery(self.lhs)
        rhs_repr = format_subquery(self.rhs)
        return f'{lhs_repr} []* {rhs_repr}'


@dataclass
class Operator(Query):
    """Class representing multiple queries joined by some operator (disjunction, for example)."""

    operator: str
    queries: list[Query]

    def referenced_identifiers(self) -> set[query.Identifier]:
        return flatmap_set(self.queries, lambda q: q.referenced_identifiers())

    def format(self, environment: Environment) -> str:
        parts = []
        for q in self.queries:
            if isinstance(q, (Token, Sequence)):
                parts.append(q.format(environment))
            else:
                parts.append('(' + q.format(environment) + ')')
        return f' {self.operator} '.join(parts)


@dataclass
class Token(Query):
    """Class representing a query for a single token."""

    identifier: query.Identifier
    # Initialize with empty set by default.
    associated_predicates: set[query.Predicate] = field(default_factory=set)
    associated_dependencies: set[query.Dependency] = field(default_factory=set)

    def referenced_identifiers(self) -> set[query.Identifier]:
        identifiers = flatmap_set(
            self.associated_dependencies, lambda r: r.referenced_identifiers()
        )
        identifiers |= flatmap_set(self.associated_predicates, lambda a: a.referenced_identifiers())
        identifiers -= {self.identifier}
        return identifiers

    def format(self, environment: Environment) -> str:
        prefix = environment[self.identifier] + ':' if self.identifier in environment else ''
        predicates = []
        for attribute in self.associated_predicates:
            # Expand conjunction as all predicates on token are already conjunct.
            if isinstance(attribute, query.Conjunction):
                predicates.extend(format_predicate(p, environment) for p in attribute.predicates)
            else:
                predicates.append(format_predicate(attribute, environment))

        for dependency in self.associated_dependencies:
            if dependency.src == self.identifier:
                dst = environment[dependency.dst]
                predicates.append(f'{dst}.dephead = ref')
            else:
                src = environment[dependency.src]
                predicates.append(f'dephead = {src}.ref')

        predicate = ' & '.join(predicates)
        return f'{prefix}[{predicate}]'


def format_operand(operand: query.Operand, environment: Environment) -> str:
    if isinstance(operand, query.Attribute):
        if operand.reference is not None:
            return f'{environment[operand.reference]}.{operand.attribute}'
        return operand.attribute
    elif isinstance(operand, query.Reference):
        if operand.reference is not None:
            return environment[operand.reference]
        return '_'
    elif isinstance(operand, query.Function):
        args = ', '.join(format_operand(arg, environment) for arg in operand.args)
        return f'{operand.name}({args})'
    assert isinstance(operand, query.Literal), 'Operand must be either Attribute or Literal.'
    return operand.value


def format_predicate(predicate: query.Predicate, environment: Environment) -> str:
    if isinstance(predicate, query.Exists):
        return format_operand(predicate.attribute, environment)
    elif isinstance(predicate, query.Negation):
        return f'!{format_predicate(predicate.predicate, environment)}'
    if isinstance(predicate, query.Comparison):
        lhs = format_operand(predicate.lhs, environment)
        rhs = format_operand(predicate.rhs, environment)
        return f'({lhs} {predicate.operator} {rhs})'
    else:
        assert isinstance(predicate, (query.Conjunction, query.Disjunction))
        operators = {
            query.Conjunction: '&',
            query.Disjunction: '|',
        }
        predicates = map(lambda p: format_predicate(p, environment), predicate.predicates)
        return to_str(predicates, '(', f' {operators[type(predicate)]} ', ')')


def arrangements(
    identifiers: set[query.Identifier],
    constraints: Iterable[query.Constraint],
) -> Iterator[list[query.Identifier]]:
    """Arrange a set of Identifiers into all sequences allowed by the given Constraints"""
    cannot_be_after = {i: set() for i in identifiers}
    for constraint in constraints:
        if isinstance(constraint, query.Constraint.Order):
            fst, snd = constraint
            cannot_be_after[snd].add(fst)

    # Buffer with space for all identifiers.
    arrangement: list[query.Identifier | None] = [None] * len(identifiers)

    def arrange(index: int, remaining_identifiers: set[query.Identifier]):
        if index == len(arrangement):
            yield list(arrangement)  # Everything is put into an order. Yield it!
        else:
            for identifier in remaining_identifiers:
                arrangement[index] = identifier

                restricted = cannot_be_after[identifier]  # Continue for remaining identifiers.
                yield from arrange(index + 1, (remaining_identifiers - restricted) - {identifier})

    yield from arrange(0, identifiers)


def from_arrangement(
    arrangement: list[query.Identifier],
    dependencies: set[query.Dependency],
    predicates: set[query.Predicate],
) -> Query:
    """Build a Query for a sequence of Identifiers."""
    visited_tokens: set[query.Identifier] = set()
    remaining_dependencies = dependencies
    remaining_predicates = predicates

    converted_tokens = [Token(i) for i in arrangement]
    for token in converted_tokens:
        visited_tokens.add(token.identifier)

        committable_dependencies, remaining_dependencies = partition_set(
            remaining_dependencies,
            lambda rel: rel.referenced_identifiers().issubset(visited_tokens),
        )
        committable_predicates, remaining_predicates = partition_set(
            remaining_predicates,
            lambda pred: pred.referenced_identifiers().issubset(visited_tokens),
        )

        token.associated_dependencies.update(committable_dependencies)
        token.associated_predicates.update(  # Lower predicates when associating with token.
            p.lower_onto(token.identifier) for p in committable_predicates
        )

    # Build all tokens into a sequence.
    res = converted_tokens[0]
    for index in range(len(arrangement) - 1):
        res = Sequence(res, converted_tokens[index + 1])
    return res


def from_all_arrangements(
    identifiers: set[query.Identifier],
    dependencies: set[query.Dependency],
    constraints: set[query.Constraint],
    predicates: set[query.Predicate],
) -> Query:
    """Build a query over all sequences of Identifiers."""
    all_arrangements = [
        from_arrangement(arrangement, dependencies, predicates)
        for arrangement in arrangements(identifiers, constraints)
    ]
    return Operator('|', all_arrangements)


ORDER_TO_OPERATOR = {
    query.Compare.EQ: '=',
    query.Compare.NE: '!=',
    query.Compare.LT: '<',
    query.Compare.GT: '>',
}


def distance_to_operand(constraint: query.Constraint.Distance) -> query.Predicate:
    fst, snd = constraint
    dist_function = query.Function(
        'distabs',
        query.Reference(fst),
        query.Reference(snd),
    )
    distance_literal = query.Literal(str(constraint.distance))
    return query.Comparison(dist_function, ORDER_TO_OPERATOR[constraint.order], distance_literal)


def from_query(q: query.Query) -> Query:
    """Translate a tree-based query into a CQP query for all different arrangements of tokens."""

    predicates = set(pred.normalize() for pred in q.predicates)
    for constraint in q.constraints:
        if isinstance(constraint, query.Constraint.Distance):
            predicates.add(distance_to_operand(constraint))

    for token in q.tokens:  # Raise local predicates to prepare re-ordering.
        if token.attributes is not None:
            raised_predicate = token.attributes.raise_from(token.identifier)
            raised_predicate = raised_predicate.normalize()
            predicates.add(raised_predicate)

    return from_all_arrangements(
        {token.identifier for token in q.tokens},
        set(q.dependencies),
        set(q.constraints),
        predicates,
    )


def format_plan(plan: query.Recipe, args: Namespace | None) -> Iterator[str]:
    span = ''
    if args is not None and args.span is not None:
        span = f' within <{args.span}>'

    environment = associate_with_names(plan.identifiers(), QUERY_ALPHABET)
    parts = plan.as_dict()

    def rec(goal: query.Identifier, include_assignment: bool = True) -> Iterator[str]:
        part = parts[goal]
        if isinstance(part, query.Operation):
            op, lhs, rhs = (
                CQP_OPERATIONS[part.operator],
                environment[part.lhs],
                environment[part.rhs],
            )
            formatted = f'{op} {lhs} {rhs};'
            yield from rec(part.lhs)
            yield from rec(part.rhs)
        else:
            formatted = str(from_query(part)) + span + ';'

        if include_assignment:
            formatted = f'{environment[goal]} = {formatted}'
        yield formatted

    yield from rec(plan.goal, include_assignment=False)



import argparse
from cqp_tree.web.server import server


def argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog='cqp-tree-web',
        description='Translate tree-style corpus queries to CQP queries.',
        add_help=False,
    )

    parser.add_argument(
        '--help',
        action='store_true',
        help='Show this message and exit.',
    )
    parser.add_argument(
        '--host',
        '-h',
        metavar='HOST',
        help='The host to listen on.',
    )
    parser.add_argument(
        '--port',
        '-p',
        metavar='PORT',
        help='The port to bind to.',
    )
    parser.add_argument(
        '--debug',
        '-d',
        action='store_true',
        help='Enable debug mode.',
    )
    return parser


def main():
    parser = argument_parser()
    args = parser.parse_args()
    if args.help:
        parser.print_help()
    else:
        server.run(
            host=args.host or 'localhost',
            port=args.port or 5000,
            debug=args.debug,
        )


if __name__ == '__main__':
    main()



from typing import List

import conllu

import cqp_tree.translation as ct

# CoNLL-U fields mapped to Språkbanken Korp attributes
FIELDS2ATTRS = {
    "form": "word",
    "lemma": "lemma",
    "upos": "pos",  # but upos (actual UD tags) may be added soon
    "xpos": "msd",  # not sure about this one
    "feats": "ufeats",  # actual UD features
    "deprel": "deprel",  # mambadep, but UD relations may be added soon
    # id and head are not treated as attributes
    # deps and misc are ignored for the time being
}


def parse(s: str):
    try:
        parsed = conllu.parse(s)
    except conllu.exceptions.ParseException as ex:
        raise ct.ParsingFailed(ct.InputError(None, ex))
    if len(parsed) == 0:
        raise ct.ParsingFailed(ct.InputError(None, 'Cannot parse an empty .conllu file.'))
    return parsed[0]  # only first parsed CoNLL-U sentence


@ct.translator('conll')
def translate_conll(conll: str) -> ct.Recipe:
    tokens: List[ct.Token] = []
    dependencies: List[ct.Dependency] = []

    conll_lines = parse(conll)

    ids = [ct.Identifier() for _ in conll_lines]

    def field2op(field, value) -> ct.Comparison:
        return ct.Comparison(ct.Attribute(None, field), '=', ct.Literal(f'"{value}"'))

    def is_empty(line, field):
        return line[field] in ["_", None]

    for line, id in list(zip(conll_lines, ids)):
        if not isinstance(line["id"], int):
            continue  # skip MWE lines
        if is_empty(line, "id") or is_empty(line, "head"):
            raise ct.NotSupported("IDs and HEADs cannot be omitted.")
        ops = []
        # pylint: disable=consider-using-dict-items
        for field in FIELDS2ATTRS:
            if not is_empty(line, field):
                ops.append(field2op(FIELDS2ATTRS[field], line[field]))
        if ops:
            tokens.append(ct.Token(id, ct.Conjunction(ops)))
        else:  # token with no attributes, only there for structural reasons
            tokens.append(ct.Token(id))

        if line["head"] != 0:
            dependencies.append(
                ct.Dependency(ids[[line["id"] for line in conll_lines].index(line["head"])], id)
            )

    query = ct.Query(tokens=tokens, dependencies=dependencies)
    return ct.Recipe.of_query(query)



import unittest

import cqp_tree.translation.query as query


class ValidationTests(unittest.TestCase):

    def test_referenced_identifiers(self):
        a = query.Identifier()
        b = query.Identifier()
        c = query.Identifier()

        tree = query.Conjunction(
            [
                query.Comparison(query.Attribute(a, 'attr_a'), '===', query.Literal('42')),
                query.Disjunction(
                    [
                        query.Exists(
                            query.Attribute(b, 'attr_b'),
                        ),
                        query.Negation(query.Exists(query.Attribute(c, 'attr_d'))),
                    ]
                ),
            ]
        )

        self.assertSetEqual({a, b, c}, tree.referenced_identifiers())

    def test_constraint_references_unknown_identifier(self):
        a = query.Identifier()
        b = query.Identifier()
        unknown = query.Identifier()
        constraints = [
            query.Constraint.order(a, unknown),
            query.Constraint.order(unknown, b),
            query.Constraint.distance(a, unknown) == 0,
            query.Constraint.distance(unknown, b) == 0,
        ]

        for constraint in constraints:
            with self.assertRaises(AssertionError):
                tokens = [
                    query.Token(a, None),
                    query.Token(b, None),
                ]
                constraints = [constraint]
                query.Query(tokens=tokens, constraints=constraints)

    def test_relation_references_unknown_identifier(self):
        with self.assertRaises(AssertionError):
            a = query.Identifier()
            b = query.Identifier()
            unknown = query.Identifier()

            tokens = [
                query.Token(a, None),
                query.Token(b, None),
            ]
            relations = [query.Dependency(a, unknown)]

            query.Query(tokens=tokens, dependencies=relations)

    def test_local_predicate_references_unknown_identifier(self):
        with self.assertRaises(AssertionError):
            a = query.Identifier()
            unknown = query.Identifier()

            tokens = [
                query.Token(a, query.Exists(query.Attribute(unknown, 'attribute'))),
            ]
            query.Query(tokens=tokens)

    def test_global_predicate_references_unknown_identifier(self):
        with self.assertRaises(AssertionError):
            a = query.Identifier()
            unknown = query.Identifier()

            tokens = [
                query.Token(a, None),
            ]
            predicates = [
                query.Comparison(query.Attribute(unknown, 'attribute'), '=', query.Literal('2'))
            ]

            query.Query(tokens=tokens, predicates=predicates)

    def test_duplicate_identifiers(self):
        with self.assertRaises(AssertionError):
            duplicate = query.Identifier()

            tokens = [
                query.Token(duplicate, None),
                query.Token(duplicate, None),
            ]
            query.Query(tokens=tokens)



from typing import Any

from flask import Flask, jsonify, request, send_from_directory

import cqp_tree
from cqp_tree import Recipe
from cqp_tree.utils import UPPERCASE_ALPHABET, associate_with_names

server = Flask(__name__)


@server.route("/")
def main():
    return send_from_directory('static', 'index.html')


@server.route('/translators', methods=['GET'])
def get_translators():
    translators = sorted(cqp_tree.known_translators.keys())
    return jsonify(translators)


@server.route('/translate', methods=['POST'])
def translate():
    def error(message: str, status: int = 400):
        return jsonify({'error': message}), status

    def extract_request_data():
        translation_request = request.get_json()
        if translation_request is None or not isinstance(translation_request, dict):
            raise ValueError('Malformed request')

        if not 'text' in translation_request:
            raise ValueError('Missing required field "text"')

        text = translation_request['text']
        translator = translation_request.get('translator')
        if translator and translator not in cqp_tree.known_translators:
            raise ValueError('Unknown value for field "translator"')
        return text, translator

    try:
        text, translator = extract_request_data()
        plan = cqp_tree.translate_input(text, translator)

        if is_too_complex(plan):
            raise ValueError('Your query is too complex! Try using fewer tokens.')

        return jsonify(to_json(plan))

    except ValueError as validation_error:
        return error(str(validation_error), 422)

    except cqp_tree.UnableToGuessTranslatorError as unable_to_guess_translator:
        if unable_to_guess_translator.no_translator_matches():
            return error(
                'This query cannot be translated. '
                'Try checking for syntax errors or manually select the query language.'
            )
        return error(
            'This query is valid in multiple query languages. '
            'Please manually select the query language you intend.'
        )

    except cqp_tree.NotSupported as not_supported:
        return error('This query is not supported: ' + str(not_supported))

    except cqp_tree.ParsingFailed as parse_error:
        parse_error = next(iter(parse_error.errors))
        return error('This query cannot be parsed: ' + parse_error.message)


def is_too_complex(plan: Recipe) -> bool:
    if len(plan.queries) > 20:
        return True
    if any(len(query.tokens) > 5 for query in plan.queries):
        return True
    return False


def to_json(plan: Recipe) -> dict:
    environment = associate_with_names(plan.identifiers(), UPPERCASE_ALPHABET)

    queries = {
        environment[query.identifier]: str(cqp_tree.cqp_from_query(query)) for query in plan.queries
    }
    operations = {
        environment[operation.identifier]: {
            'lhs': environment[operation.lhs],
            'rhs': environment[operation.rhs],
            'op': operation.operator,
        }
        for operation in plan.operations
    }

    result: dict[str, Any] = {
        'recipe': {
            'queries': queries,
            'operations': operations,
            'goal': environment[plan.goal],
        }
    }
    if plan.has_simple_representation():
        result['single_query'] = str(cqp_tree.cqp_from_query(plan.simple_representation()))

    return result



import argparse
import sys
from contextlib import ExitStack
from typing import Any, Optional

import cqp_tree
from cqp_tree.utils import format_human_readable


def warn(msg: Any):
    print(msg, file=sys.stderr)


def argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog='cqp-tree',
        description='Translate tree-style corpus queries to CQP queries.',
        add_help=False,
    )

    parser.add_argument(
        '--help',
        '-h',
        action='store_true',
        help='Show this message and exit.',
    )
    parser.add_argument(
        '--quiet',
        action='store_true',
        help='Don\'t print a message when reading a query from standard input.',
    )
    parser.add_argument(
        '--output',
        '-o',
        metavar='FILE',
        help='Output file to which results are written. '
        'If omitted, results are printed to stdout instead.',
    )
    parser.add_argument(
        '--encoding',
        '-e',
        default='utf-8',
        metavar='ENC',
        help='Encoding used for reading and writing files.',
    )
    parser.add_argument(
        '--span',
        '-s',
        metavar='SPAN',
        help='Span attribute to which a query should be constrained.',
    )

    translator_names = sorted(cqp_tree.known_translators.keys())
    parser.add_argument(
        'translator',
        metavar='TRANSLATOR',
        help='The translator to choose. '
        'If not provided, a translator is determined automatically for each query. '
        'Supported options are: ' + ', '.join(translator_names),
        nargs='?',
        choices=translator_names,
    )

    input_group = parser.add_mutually_exclusive_group()
    input_group.add_argument(
        '--file',
        '-f',
        metavar='FILE',
        help='Input file containing a query to translate.',
    )
    input_group.add_argument(
        '--query',
        '-q',
        metavar='STR',
        help='A query to translate.',
    )

    return parser


def get_input(args: argparse.Namespace) -> Optional[str]:
    if args.file:
        try:
            with open(args.file, 'r', encoding=args.encoding) as f:
                return f.read()
        except IOError as e:
            warn(f'Could not read input file {args.file}: {e}')
            return None

    elif args.query:
        return args.query

    else:
        if not args.quiet:
            warn('No input file specified. Reading from stdin instead.')
            warn('Press Ctrl+D once you finished typing your query.')
        return sys.stdin.read() or None


def translate(args: argparse.Namespace, query_str: str) -> cqp_tree.Recipe | None:
    try:
        return cqp_tree.translate_input(query_str, args.translator or None)
    except cqp_tree.UnableToGuessTranslatorError as translation_error:
        if translation_error.no_translator_matches():
            warn('Unable to determine translator: No translator accepts the query.')
        else:
            accepting_translators = format_human_readable(
                sorted(translation_error.matching_translators)
            )
            warn(f'Unable to determine translator: Query is accepted by {accepting_translators}')
    return None


def main():
    parser = argument_parser()
    args = parser.parse_args()
    if args.help:
        parser.print_help()
        return 0

    with ExitStack() as managed_resources:
        output = sys.stdout
        if args.output:
            try:
                output = managed_resources.enter_context(
                    open(args.output, 'w', encoding=args.encoding)
                )
            except IOError as e:
                warn(f'Could not write to output file {args.output}: {e}')
                return 1

        query_str = get_input(args)
        if query_str is None:
            return 1

        try:
            plan = translate(args, query_str)
            if not plan:
                return 1

            for line in cqp_tree.format_plan(plan, args):
                output.write(line + '\n')

        except cqp_tree.ParsingFailed as parse_failure:
            warn('Query could not be parsed:')
            for error in parse_failure.errors:
                warn(error)
        except cqp_tree.NotSupported as not_supported:
            if not str(not_supported):
                warn('Query cannot be translated.')
            else:
                warn('Query cannot be translated: ' + str(not_supported))

        return 0


if __name__ == '__main__':
    sys.exit(main())



# Generated from Grew.g4 by ANTLR 4.13.2
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,27,163,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,2,9,7,9,2,10,7,10,2,11,7,11,2,12,7,12,1,0,1,0,
        5,0,29,8,0,10,0,12,0,32,9,0,1,0,1,0,1,1,1,1,1,1,1,2,1,2,1,2,1,2,
        3,2,43,8,2,1,3,1,3,1,3,1,3,5,3,49,8,3,10,3,12,3,52,9,3,1,3,1,3,3,
        3,56,8,3,3,3,58,8,3,1,3,1,3,1,4,1,4,1,4,1,4,5,4,66,8,4,10,4,12,4,
        69,9,4,1,4,1,4,3,4,73,8,4,1,4,1,4,1,4,1,4,1,4,1,4,1,4,1,4,1,4,1,
        4,1,4,1,4,3,4,87,8,4,1,5,1,5,1,5,1,5,5,5,93,8,5,10,5,12,5,96,9,5,
        3,5,98,8,5,1,5,1,5,1,6,1,6,1,6,1,6,1,6,1,6,1,6,1,6,5,6,110,8,6,10,
        6,12,6,113,9,6,3,6,115,8,6,3,6,117,8,6,1,7,1,7,1,7,1,7,1,7,1,7,1,
        7,1,7,1,7,1,7,3,7,129,8,7,1,8,1,8,1,8,5,8,134,8,8,10,8,12,8,137,
        9,8,1,9,1,9,1,9,1,9,3,9,143,8,9,1,10,1,10,1,10,1,10,1,10,1,10,1,
        10,1,10,3,10,153,8,10,1,11,1,11,3,11,157,8,11,1,12,1,12,3,12,161,
        8,12,1,12,0,0,13,0,2,4,6,8,10,12,14,16,18,20,22,24,0,0,175,0,26,
        1,0,0,0,2,35,1,0,0,0,4,42,1,0,0,0,6,44,1,0,0,0,8,86,1,0,0,0,10,88,
        1,0,0,0,12,116,1,0,0,0,14,128,1,0,0,0,16,130,1,0,0,0,18,142,1,0,
        0,0,20,152,1,0,0,0,22,156,1,0,0,0,24,160,1,0,0,0,26,30,3,2,1,0,27,
        29,3,4,2,0,28,27,1,0,0,0,29,32,1,0,0,0,30,28,1,0,0,0,30,31,1,0,0,
        0,31,33,1,0,0,0,32,30,1,0,0,0,33,34,5,0,0,1,34,1,1,0,0,0,35,36,5,
        1,0,0,36,37,3,6,3,0,37,3,1,0,0,0,38,39,5,2,0,0,39,43,3,6,3,0,40,
        41,5,3,0,0,41,43,3,6,3,0,42,38,1,0,0,0,42,40,1,0,0,0,43,5,1,0,0,
        0,44,50,5,4,0,0,45,46,3,8,4,0,46,47,5,5,0,0,47,49,1,0,0,0,48,45,
        1,0,0,0,49,52,1,0,0,0,50,48,1,0,0,0,50,51,1,0,0,0,51,57,1,0,0,0,
        52,50,1,0,0,0,53,55,3,8,4,0,54,56,5,5,0,0,55,54,1,0,0,0,55,56,1,
        0,0,0,56,58,1,0,0,0,57,53,1,0,0,0,57,58,1,0,0,0,58,59,1,0,0,0,59,
        60,5,6,0,0,60,7,1,0,0,0,61,62,5,25,0,0,62,67,3,10,5,0,63,64,5,7,
        0,0,64,66,3,10,5,0,65,63,1,0,0,0,66,69,1,0,0,0,67,65,1,0,0,0,67,
        68,1,0,0,0,68,87,1,0,0,0,69,67,1,0,0,0,70,71,5,25,0,0,71,73,5,8,
        0,0,72,70,1,0,0,0,72,73,1,0,0,0,73,74,1,0,0,0,74,75,5,25,0,0,75,
        76,3,14,7,0,76,77,5,25,0,0,77,87,1,0,0,0,78,79,3,18,9,0,79,80,3,
        24,12,0,80,81,3,18,9,0,81,87,1,0,0,0,82,83,5,25,0,0,83,84,3,22,11,
        0,84,85,5,25,0,0,85,87,1,0,0,0,86,61,1,0,0,0,86,72,1,0,0,0,86,78,
        1,0,0,0,86,82,1,0,0,0,87,9,1,0,0,0,88,97,5,9,0,0,89,94,3,12,6,0,
        90,91,5,10,0,0,91,93,3,12,6,0,92,90,1,0,0,0,93,96,1,0,0,0,94,92,
        1,0,0,0,94,95,1,0,0,0,95,98,1,0,0,0,96,94,1,0,0,0,97,89,1,0,0,0,
        97,98,1,0,0,0,98,99,1,0,0,0,99,100,5,11,0,0,100,11,1,0,0,0,101,117,
        5,25,0,0,102,103,5,12,0,0,103,117,5,25,0,0,104,105,5,25,0,0,105,
        114,3,24,12,0,106,111,3,18,9,0,107,108,5,7,0,0,108,110,3,18,9,0,
        109,107,1,0,0,0,110,113,1,0,0,0,111,109,1,0,0,0,111,112,1,0,0,0,
        112,115,1,0,0,0,113,111,1,0,0,0,114,106,1,0,0,0,114,115,1,0,0,0,
        115,117,1,0,0,0,116,101,1,0,0,0,116,102,1,0,0,0,116,104,1,0,0,0,
        117,13,1,0,0,0,118,129,5,13,0,0,119,120,5,14,0,0,120,121,3,16,8,
        0,121,122,5,15,0,0,122,129,1,0,0,0,123,124,5,14,0,0,124,125,5,16,
        0,0,125,126,3,16,8,0,126,127,5,15,0,0,127,129,1,0,0,0,128,118,1,
        0,0,0,128,119,1,0,0,0,128,123,1,0,0,0,129,15,1,0,0,0,130,135,3,20,
        10,0,131,132,5,7,0,0,132,134,3,20,10,0,133,131,1,0,0,0,134,137,1,
        0,0,0,135,133,1,0,0,0,135,136,1,0,0,0,136,17,1,0,0,0,137,135,1,0,
        0,0,138,139,5,25,0,0,139,140,5,17,0,0,140,143,5,25,0,0,141,143,3,
        20,10,0,142,138,1,0,0,0,142,141,1,0,0,0,143,19,1,0,0,0,144,153,5,
        26,0,0,145,153,5,25,0,0,146,147,5,18,0,0,147,153,5,26,0,0,148,153,
        5,27,0,0,149,150,5,25,0,0,150,151,5,8,0,0,151,153,5,25,0,0,152,144,
        1,0,0,0,152,145,1,0,0,0,152,146,1,0,0,0,152,148,1,0,0,0,152,149,
        1,0,0,0,153,21,1,0,0,0,154,157,5,19,0,0,155,157,5,20,0,0,156,154,
        1,0,0,0,156,155,1,0,0,0,157,23,1,0,0,0,158,161,5,21,0,0,159,161,
        5,22,0,0,160,158,1,0,0,0,160,159,1,0,0,0,161,25,1,0,0,0,19,30,42,
        50,55,57,67,72,86,94,97,111,114,116,128,135,142,152,156,160
    ]

class GrewParser ( Parser ):

    grammarFileName = "Grew.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'pattern'", "'with'", "'without'", "'{'", 
                     "';'", "'}'", "'|'", "':'", "'['", "','", "']'", "'!'", 
                     "'->'", "'-['", "']->'", "'^'", "'.'", "'re'", "'<'", 
                     "'<<'", "'='", "'<>'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "IgnoredWhitespace", 
                      "IgnoredComments", "Identifier", "String", "PCREString" ]

    RULE_request = 0
    RULE_pattern = 1
    RULE_requestItem = 2
    RULE_body = 3
    RULE_clause = 4
    RULE_featureStructure = 5
    RULE_feature = 6
    RULE_arrow = 7
    RULE_edgeTypes = 8
    RULE_featureValue = 9
    RULE_literal = 10
    RULE_order = 11
    RULE_compare = 12

    ruleNames =  [ "request", "pattern", "requestItem", "body", "clause", 
                   "featureStructure", "feature", "arrow", "edgeTypes", 
                   "featureValue", "literal", "order", "compare" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    T__10=11
    T__11=12
    T__12=13
    T__13=14
    T__14=15
    T__15=16
    T__16=17
    T__17=18
    T__18=19
    T__19=20
    T__20=21
    T__21=22
    IgnoredWhitespace=23
    IgnoredComments=24
    Identifier=25
    String=26
    PCREString=27

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class RequestContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def pattern(self):
            return self.getTypedRuleContext(GrewParser.PatternContext,0)


        def EOF(self):
            return self.getToken(GrewParser.EOF, 0)

        def requestItem(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GrewParser.RequestItemContext)
            else:
                return self.getTypedRuleContext(GrewParser.RequestItemContext,i)


        def getRuleIndex(self):
            return GrewParser.RULE_request




    def request(self):

        localctx = GrewParser.RequestContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_request)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 26
            self.pattern()
            self.state = 30
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==2 or _la==3:
                self.state = 27
                self.requestItem()
                self.state = 32
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 33
            self.match(GrewParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PatternContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def body(self):
            return self.getTypedRuleContext(GrewParser.BodyContext,0)


        def getRuleIndex(self):
            return GrewParser.RULE_pattern




    def pattern(self):

        localctx = GrewParser.PatternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_pattern)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 35
            self.match(GrewParser.T__0)
            self.state = 36
            self.body()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class RequestItemContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return GrewParser.RULE_requestItem

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class WithItemContext(RequestItemContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.RequestItemContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def body(self):
            return self.getTypedRuleContext(GrewParser.BodyContext,0)



    class WithoutItemContext(RequestItemContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.RequestItemContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def body(self):
            return self.getTypedRuleContext(GrewParser.BodyContext,0)




    def requestItem(self):

        localctx = GrewParser.RequestItemContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_requestItem)
        try:
            self.state = 42
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [2]:
                localctx = GrewParser.WithItemContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 38
                self.match(GrewParser.T__1)
                self.state = 39
                self.body()
                pass
            elif token in [3]:
                localctx = GrewParser.WithoutItemContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 40
                self.match(GrewParser.T__2)
                self.state = 41
                self.body()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class BodyContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def clause(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GrewParser.ClauseContext)
            else:
                return self.getTypedRuleContext(GrewParser.ClauseContext,i)


        def getRuleIndex(self):
            return GrewParser.RULE_body




    def body(self):

        localctx = GrewParser.BodyContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_body)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 44
            self.match(GrewParser.T__3)
            self.state = 50
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,2,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 45
                    self.clause()
                    self.state = 46
                    self.match(GrewParser.T__4) 
                self.state = 52
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,2,self._ctx)

            self.state = 57
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 235143168) != 0):
                self.state = 53
                self.clause()
                self.state = 55
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==5:
                    self.state = 54
                    self.match(GrewParser.T__4)




            self.state = 59
            self.match(GrewParser.T__5)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ClauseContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return GrewParser.RULE_clause

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class ConstraintClauseContext(ClauseContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.ClauseContext
            super().__init__(parser)
            self.lhs = None # FeatureValueContext
            self.rhs = None # FeatureValueContext
            self.copyFrom(ctx)

        def compare(self):
            return self.getTypedRuleContext(GrewParser.CompareContext,0)

        def featureValue(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GrewParser.FeatureValueContext)
            else:
                return self.getTypedRuleContext(GrewParser.FeatureValueContext,i)



    class EdgeClauseContext(ClauseContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.ClauseContext
            super().__init__(parser)
            self.label = None # Token
            self.src = None # Token
            self.dst = None # Token
            self.copyFrom(ctx)

        def arrow(self):
            return self.getTypedRuleContext(GrewParser.ArrowContext,0)

        def Identifier(self, i:int=None):
            if i is None:
                return self.getTokens(GrewParser.Identifier)
            else:
                return self.getToken(GrewParser.Identifier, i)


    class OrderClauseContext(ClauseContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.ClauseContext
            super().__init__(parser)
            self.lhs = None # Token
            self.rhs = None # Token
            self.copyFrom(ctx)

        def order(self):
            return self.getTypedRuleContext(GrewParser.OrderContext,0)

        def Identifier(self, i:int=None):
            if i is None:
                return self.getTokens(GrewParser.Identifier)
            else:
                return self.getToken(GrewParser.Identifier, i)


    class NodeClauseContext(ClauseContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.ClauseContext
            super().__init__(parser)
            self.label = None # Token
            self.copyFrom(ctx)

        def featureStructure(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GrewParser.FeatureStructureContext)
            else:
                return self.getTypedRuleContext(GrewParser.FeatureStructureContext,i)

        def Identifier(self):
            return self.getToken(GrewParser.Identifier, 0)



    def clause(self):

        localctx = GrewParser.ClauseContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_clause)
        self._la = 0 # Token type
        try:
            self.state = 86
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,7,self._ctx)
            if la_ == 1:
                localctx = GrewParser.NodeClauseContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 61
                localctx.label = self.match(GrewParser.Identifier)
                self.state = 62
                self.featureStructure()
                self.state = 67
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==7:
                    self.state = 63
                    self.match(GrewParser.T__6)
                    self.state = 64
                    self.featureStructure()
                    self.state = 69
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                pass

            elif la_ == 2:
                localctx = GrewParser.EdgeClauseContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 72
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,6,self._ctx)
                if la_ == 1:
                    self.state = 70
                    localctx.label = self.match(GrewParser.Identifier)
                    self.state = 71
                    self.match(GrewParser.T__7)


                self.state = 74
                localctx.src = self.match(GrewParser.Identifier)
                self.state = 75
                self.arrow()
                self.state = 76
                localctx.dst = self.match(GrewParser.Identifier)
                pass

            elif la_ == 3:
                localctx = GrewParser.ConstraintClauseContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 78
                localctx.lhs = self.featureValue()
                self.state = 79
                self.compare()
                self.state = 80
                localctx.rhs = self.featureValue()
                pass

            elif la_ == 4:
                localctx = GrewParser.OrderClauseContext(self, localctx)
                self.enterOuterAlt(localctx, 4)
                self.state = 82
                localctx.lhs = self.match(GrewParser.Identifier)
                self.state = 83
                self.order()
                self.state = 84
                localctx.rhs = self.match(GrewParser.Identifier)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class FeatureStructureContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def feature(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GrewParser.FeatureContext)
            else:
                return self.getTypedRuleContext(GrewParser.FeatureContext,i)


        def getRuleIndex(self):
            return GrewParser.RULE_featureStructure




    def featureStructure(self):

        localctx = GrewParser.FeatureStructureContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_featureStructure)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 88
            self.match(GrewParser.T__8)
            self.state = 97
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==12 or _la==25:
                self.state = 89
                self.feature()
                self.state = 94
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==10:
                    self.state = 90
                    self.match(GrewParser.T__9)
                    self.state = 91
                    self.feature()
                    self.state = 96
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)



            self.state = 99
            self.match(GrewParser.T__10)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class FeatureContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return GrewParser.RULE_feature

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class RequiresContext(FeatureContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.FeatureContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def Identifier(self):
            return self.getToken(GrewParser.Identifier, 0)
        def compare(self):
            return self.getTypedRuleContext(GrewParser.CompareContext,0)

        def featureValue(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GrewParser.FeatureValueContext)
            else:
                return self.getTypedRuleContext(GrewParser.FeatureValueContext,i)



    class PresenceContext(FeatureContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.FeatureContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def Identifier(self):
            return self.getToken(GrewParser.Identifier, 0)


    class AbsenceContext(FeatureContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.FeatureContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def Identifier(self):
            return self.getToken(GrewParser.Identifier, 0)



    def feature(self):

        localctx = GrewParser.FeatureContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_feature)
        self._la = 0 # Token type
        try:
            self.state = 116
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,12,self._ctx)
            if la_ == 1:
                localctx = GrewParser.PresenceContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 101
                self.match(GrewParser.Identifier)
                pass

            elif la_ == 2:
                localctx = GrewParser.AbsenceContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 102
                self.match(GrewParser.T__11)
                self.state = 103
                self.match(GrewParser.Identifier)
                pass

            elif la_ == 3:
                localctx = GrewParser.RequiresContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 104
                self.match(GrewParser.Identifier)
                self.state = 105
                self.compare()
                self.state = 114
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & 235143168) != 0):
                    self.state = 106
                    self.featureValue()
                    self.state = 111
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    while _la==7:
                        self.state = 107
                        self.match(GrewParser.T__6)
                        self.state = 108
                        self.featureValue()
                        self.state = 113
                        self._errHandler.sync(self)
                        _la = self._input.LA(1)



                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ArrowContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return GrewParser.RULE_arrow

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class PositiveArrowContext(ArrowContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.ArrowContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def edgeTypes(self):
            return self.getTypedRuleContext(GrewParser.EdgeTypesContext,0)



    class NegatedArrowContext(ArrowContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.ArrowContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def edgeTypes(self):
            return self.getTypedRuleContext(GrewParser.EdgeTypesContext,0)



    class SimpleArrowContext(ArrowContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.ArrowContext
            super().__init__(parser)
            self.copyFrom(ctx)




    def arrow(self):

        localctx = GrewParser.ArrowContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_arrow)
        try:
            self.state = 128
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,13,self._ctx)
            if la_ == 1:
                localctx = GrewParser.SimpleArrowContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 118
                self.match(GrewParser.T__12)
                pass

            elif la_ == 2:
                localctx = GrewParser.PositiveArrowContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 119
                self.match(GrewParser.T__13)
                self.state = 120
                self.edgeTypes()
                self.state = 121
                self.match(GrewParser.T__14)
                pass

            elif la_ == 3:
                localctx = GrewParser.NegatedArrowContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 123
                self.match(GrewParser.T__13)
                self.state = 124
                self.match(GrewParser.T__15)
                self.state = 125
                self.edgeTypes()
                self.state = 126
                self.match(GrewParser.T__14)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class EdgeTypesContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def literal(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GrewParser.LiteralContext)
            else:
                return self.getTypedRuleContext(GrewParser.LiteralContext,i)


        def getRuleIndex(self):
            return GrewParser.RULE_edgeTypes




    def edgeTypes(self):

        localctx = GrewParser.EdgeTypesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_edgeTypes)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 130
            self.literal()
            self.state = 135
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==7:
                self.state = 131
                self.match(GrewParser.T__6)
                self.state = 132
                self.literal()
                self.state = 137
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class FeatureValueContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return GrewParser.RULE_featureValue

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class AttributeContext(FeatureValueContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.FeatureValueContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def Identifier(self, i:int=None):
            if i is None:
                return self.getTokens(GrewParser.Identifier)
            else:
                return self.getToken(GrewParser.Identifier, i)


    class ValueContext(FeatureValueContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.FeatureValueContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def literal(self):
            return self.getTypedRuleContext(GrewParser.LiteralContext,0)




    def featureValue(self):

        localctx = GrewParser.FeatureValueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_featureValue)
        try:
            self.state = 142
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,15,self._ctx)
            if la_ == 1:
                localctx = GrewParser.AttributeContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 138
                self.match(GrewParser.Identifier)
                self.state = 139
                self.match(GrewParser.T__16)
                self.state = 140
                self.match(GrewParser.Identifier)
                pass

            elif la_ == 2:
                localctx = GrewParser.ValueContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 141
                self.literal()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class LiteralContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return GrewParser.RULE_literal

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class PCREContext(LiteralContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.LiteralContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def PCREString(self):
            return self.getToken(GrewParser.PCREString, 0)


    class RegexContext(LiteralContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.LiteralContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def String(self):
            return self.getToken(GrewParser.String, 0)


    class UnicodeStringContext(LiteralContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.LiteralContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def String(self):
            return self.getToken(GrewParser.String, 0)


    class SimpleStringContext(LiteralContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.LiteralContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def Identifier(self):
            return self.getToken(GrewParser.Identifier, 0)


    class SubtypeContext(LiteralContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.LiteralContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def Identifier(self, i:int=None):
            if i is None:
                return self.getTokens(GrewParser.Identifier)
            else:
                return self.getToken(GrewParser.Identifier, i)



    def literal(self):

        localctx = GrewParser.LiteralContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_literal)
        try:
            self.state = 152
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,16,self._ctx)
            if la_ == 1:
                localctx = GrewParser.UnicodeStringContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 144
                self.match(GrewParser.String)
                pass

            elif la_ == 2:
                localctx = GrewParser.SimpleStringContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 145
                self.match(GrewParser.Identifier)
                pass

            elif la_ == 3:
                localctx = GrewParser.RegexContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 146
                self.match(GrewParser.T__17)
                self.state = 147
                self.match(GrewParser.String)
                pass

            elif la_ == 4:
                localctx = GrewParser.PCREContext(self, localctx)
                self.enterOuterAlt(localctx, 4)
                self.state = 148
                self.match(GrewParser.PCREString)
                pass

            elif la_ == 5:
                localctx = GrewParser.SubtypeContext(self, localctx)
                self.enterOuterAlt(localctx, 5)
                self.state = 149
                self.match(GrewParser.Identifier)
                self.state = 150
                self.match(GrewParser.T__7)
                self.state = 151
                self.match(GrewParser.Identifier)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class OrderContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return GrewParser.RULE_order

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class PrecedenceContext(OrderContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.OrderContext
            super().__init__(parser)
            self.copyFrom(ctx)



    class ImmediatePrecedenceContext(OrderContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.OrderContext
            super().__init__(parser)
            self.copyFrom(ctx)




    def order(self):

        localctx = GrewParser.OrderContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_order)
        try:
            self.state = 156
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [19]:
                localctx = GrewParser.ImmediatePrecedenceContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 154
                self.match(GrewParser.T__18)
                pass
            elif token in [20]:
                localctx = GrewParser.PrecedenceContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 155
                self.match(GrewParser.T__19)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CompareContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return GrewParser.RULE_compare

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class EqualityContext(CompareContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.CompareContext
            super().__init__(parser)
            self.copyFrom(ctx)



    class InequalityContext(CompareContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a GrewParser.CompareContext
            super().__init__(parser)
            self.copyFrom(ctx)




    def compare(self):

        localctx = GrewParser.CompareContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_compare)
        try:
            self.state = 160
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [21]:
                localctx = GrewParser.EqualityContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 158
                self.match(GrewParser.T__20)
                pass
            elif token in [22]:
                localctx = GrewParser.InequalityContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 159
                self.match(GrewParser.T__21)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








import importlib
import os

# All sub-directories in here are expected to be a module holding a frontend.
# As part of the frontend module, there should be a @translator decorated function.
# Its definition will also register the frontend for use elsewhere.


def load_submodules():
    directory = os.path.dirname(__file__)

    for frontend in os.listdir(directory):
        if os.path.isdir(os.path.join(directory, frontend)):
            importlib.import_module(f'.{frontend}', __package__)


load_submodules()



import unittest

import cqp_tree.translation.query as query


class SyntaxTests(unittest.TestCase):
    a = query.Identifier()
    b = query.Identifier()

    def test_order_constraint(self):
        constraint = query.Constraint.order(self.a, self.b)
        self.assertIsInstance(constraint, query.Constraint)

    def test_distance_constraint(self):
        constraints = [
            query.Constraint.distance(self.a, self.b) < 1,
            query.Constraint.distance(self.a, self.b) <= 1,
            query.Constraint.distance(self.a, self.b) > 1,
            query.Constraint.distance(self.a, self.b) >= 1,
            query.Constraint.distance(self.a, self.b) == 1,
            query.Constraint.distance(self.a, self.b) != 1,
        ]

        for constraint in constraints:
            self.assertIsInstance(constraint, query.Constraint)



from cqp_tree.translation.cqp import format_plan, from_query as cqp_from_query
from cqp_tree.translation.errors import InputError, ParsingFailed, NotSupported
from cqp_tree.translation.query import (
    Identifier,
    Token,
    # Operand + subclasses
    Operand,
    Literal,
    Attribute,
    # Predicate + subclasses
    Predicate,
    Comparison,
    Exists,
    Negation,
    Conjunction,
    Disjunction,
    # Query structure
    Dependency,
    Constraint,
    Compare,
    # Multi-part queries
    Query,
    Operation,
    SetOperator,
    Recipe,
)
from cqp_tree.translation.registry import (
    UnableToGuessTranslatorError,
    known_translators,
    translate_input,
    translator,
)



from dataclasses import dataclass
from typing import Callable, Collection, Optional, Tuple

from cqp_tree.translation.errors import NotSupported
from cqp_tree.translation.query import Recipe

type TranslationFunction = Callable[[str], Recipe]

known_translators = dict[str, TranslationFunction]()


def translator(name: str):
    """
    Decorator used to register translation functions.
    """

    def register(func: TranslationFunction) -> TranslationFunction:
        if name in known_translators:
            raise ValueError(f'Another translation function for {name} has already been registered')
        known_translators[name] = func
        return func

    return register


@dataclass(frozen=True)
class UnableToGuessTranslatorError(Exception):
    matching_translators: Collection[str]

    def no_translator_matches(self) -> bool:
        return not self.matching_translators

    def too_many_translators_match(self) -> bool:
        return len(self.matching_translators) > 1

    def __str__(self):
        if self.no_translator_matches():
            reason = 'no translator matches'
        else:
            reason = 'multiple translators match'
        return f'Cannot guess translator for query: {reason}'


def translate_input(inp: str, use_translator: Optional[str] = None) -> Recipe:
    """
    Translates an input using the given translator. If no translator is given,
    the correct translator is guessed by trying all available translators.

    If guessing the translator gives 0 or multiple possible translators, an
    UnableToGuessTranslatorError is raised, containing all applicable translators.

    If a translator to use is specified, but the translator is not known,
    a KeyError is raised.
    """
    if use_translator is None:
        guessed_translations = guess_correct_translator(inp)
        if not guessed_translations:
            raise UnableToGuessTranslatorError(tuple())

        if len(guessed_translations) > 1:
            raise UnableToGuessTranslatorError(tuple(trans for trans, _ in guessed_translations))

        _, query = guessed_translations[0]
        return query

    if use_translator not in known_translators:
        raise KeyError(f'Unknown translator: {use_translator}')
    return known_translators[use_translator](inp)


def guess_correct_translator(inp: str) -> list[Tuple[str, Recipe]]:
    """
    Tries to find translators applicable for the input string.
    Returns all successfully translated queries and the name of the translation frontend that
    accepted the input.

    If only one translator parses the query but raises a NotSupported, this exception is propagated.

    :param inp: The input for which translation is attempted by all frontends.
    """
    translated_queries = list[Tuple[str, Recipe]]()
    unsupported_queries = list[Tuple[str, NotSupported]]()

    for name, function in known_translators.items():
        try:
            parsed = function(inp)
            translated_queries.append((name, parsed))
        except NotSupported as not_supported:
            unsupported_queries.append((name, not_supported))
        except:  # pylint: disable=bare-except
            pass  # Assume that we cannot translate, independent of exception raised.

    if not translated_queries and len(unsupported_queries) == 1:
        trans, raised_exception = unsupported_queries[0]
        # Create a new exception with copied message and include original traceback.
        raise NotSupported(
            f'{raised_exception} (automatically selected {trans} as a translator)'
        ) from raised_exception
    return translated_queries



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import StrEnum
from itertools import count
from typing import ClassVar, Collection, Iterable, List, Optional, Self, Set, override

from cqp_tree.translation.regex import escape_regex_string
from cqp_tree.utils import flatmap_set


class Identifier:
    _ids = count(0)

    def __init__(self):
        self.id = next(Identifier._ids)

    def __repr__(self):
        return f'Identifier({self.id})'


class Operand(ABC):
    """
    Abstract superclass for values within a Predicate.
    """

    @abstractmethod
    def referenced_identifiers(self) -> set[Identifier]:
        """
        Returns a set of all identifiers appearing in this Operand.
        """

    @abstractmethod
    def raise_from(self, on: Identifier) -> 'Operand':
        """
        Raise this Operand into a global context.

        This is done by explicitly introducing an identifier for "local" attributes,
        which had no identifier before.
        """

    @abstractmethod
    def lower_onto(self, on: Identifier) -> 'Operand':
        """
        Lower this Operand into a local context.

        This is done by explicitly removing identifiers to introduce "local" attributes.
        """


@dataclass(frozen=True)
class Literal(Operand):
    """An Operand representing a literal value. It's "value" field is used without modification."""

    value: str

    def __init__(self, value: str, represents_regex: bool = False):
        """
        Construct a new Literal with the given representation.

        :param value: The string used to represent this Literal.
        :param represents_regex: If true, regular expression characters will be escaped
        (default False).
        """
        if not represents_regex:
            value = escape_regex_string(value)
        super().__setattr__('value', value)

    def referenced_identifiers(self) -> set[Identifier]:
        return set()

    def raise_from(self, on: Identifier) -> 'Literal':
        return self

    def lower_onto(self, on: Identifier) -> 'Literal':
        return self


@dataclass(frozen=True)
class Reference(Operand):
    """
    An Operand representing an Identifier.
    If constructed using None, it represents the current label.
    """

    reference: Optional[Identifier]

    def referenced_identifiers(self) -> set[Identifier]:
        if self.reference is None:
            return set()
        return {self.reference}

    def raise_from(self, on: Identifier) -> 'Operand':
        if not self.reference is None:
            return Reference(on)
        return self

    def lower_onto(self, on: Identifier) -> 'Operand':
        if self.reference == on:
            return Reference(None)
        return self


@dataclass(frozen=True)
class Function(Operand):
    """An Operand applying a builtin function."""

    name: str
    args: Iterable[Operand]

    def __init__(self, name: str, *args: Operand):
        """
        Construct a new Function, using the given name and arguments.
        """
        super().__setattr__('name', name)
        super().__setattr__('args', tuple(args))

    def referenced_identifiers(self) -> set[Identifier]:
        return flatmap_set(self.args, lambda o: o.referenced_identifiers())

    def raise_from(self, on: Identifier) -> 'Operand':
        return Function(self.name, *[arg.raise_from(on) for arg in self.args])

    def lower_onto(self, on: Identifier) -> 'Operand':
        return Function(self.name, *[arg.lower_onto(on) for arg in self.args])


@dataclass(frozen=True)
class Attribute(Operand):
    """
    An Operand representing a field on a token.
    If the token reference is None, a field on the current token is referenced.
    """

    reference: Optional[Identifier]
    attribute: str

    def referenced_identifiers(self) -> set[Identifier]:
        return {self.reference} if self.reference else set()

    def raise_from(self, on: Identifier) -> 'Attribute':
        if self.reference is None:
            return Attribute(on, self.attribute)
        return self

    def lower_onto(self, on: Identifier) -> 'Attribute':
        if self.reference == on:
            return Attribute(None, self.attribute)
        return self


class Predicate(ABC):
    """Abstract superclass for all Predicates on a token."""

    @abstractmethod
    def referenced_identifiers(self) -> set[Identifier]: ...

    @abstractmethod
    def raise_from(self, on: Identifier) -> 'Predicate':
        """
        Raise this Predicate into a global context.
        """

    @abstractmethod
    def lower_onto(self, on: Identifier) -> 'Predicate':
        """
        Lower this Predicate into a local context.
        """

    @abstractmethod
    def normalize(self) -> 'Predicate':
        """
        Turns this Predicate into a simplified, immutable, hashable copy of itself.
        """


@dataclass(frozen=True)
class Comparison(Predicate):
    """A Predicate comparing two Operands using an arbitrary operator."""

    lhs: Operand
    operator: str
    rhs: Operand

    def referenced_identifiers(self) -> set[Identifier]:
        return flatmap_set([self.lhs, self.rhs], lambda o: o.referenced_identifiers())

    def raise_from(self, on: Identifier) -> 'Comparison':
        lhs = self.lhs.raise_from(on)
        rhs = self.rhs.raise_from(on)
        return Comparison(lhs, self.operator, rhs)

    def lower_onto(self, on: Identifier) -> 'Comparison':
        lhs = self.lhs.lower_onto(on)
        rhs = self.rhs.lower_onto(on)
        return Comparison(lhs, self.operator, rhs)

    def normalize(self) -> 'Comparison':
        return Comparison(self.lhs, self.operator, self.rhs)


@dataclass(frozen=True)
class Exists(Predicate):
    """A Predicate requiring the existence of an Attribute."""

    attribute: Attribute

    def referenced_identifiers(self) -> set[Identifier]:
        return self.attribute.referenced_identifiers()

    def raise_from(self, on: Identifier) -> 'Exists':
        return Exists(self.attribute.raise_from(on))

    def lower_onto(self, on: Identifier) -> 'Exists':
        return Exists(self.attribute.lower_onto(on))

    def normalize(self) -> 'Exists':
        return Exists(self.attribute)


@dataclass(frozen=True)
class Negation(Predicate):
    """A negated Predicate."""

    predicate: Predicate

    def referenced_identifiers(self) -> set[Identifier]:
        return self.predicate.referenced_identifiers()

    def raise_from(self, on: Identifier) -> 'Negation':
        return Negation(self.predicate.raise_from(on))

    def lower_onto(self, on: Identifier) -> 'Negation':
        return Negation(self.predicate.lower_onto(on))

    def normalize(self) -> Predicate:
        predicate = self.predicate.normalize()
        if isinstance(predicate, Negation):
            return predicate  # remove double negation.
        return Negation(predicate)


@dataclass(frozen=True)
class GenericJunction(Predicate, ABC):
    """Abstract superclass for Conjunction and Disjunction.
    Implements all their method in a generic manner."""

    predicates: Iterable[Predicate]

    def __init_subclass__(cls):
        super().__init_subclass__()
        if cls.__name__ not in {'Conjunction', 'Disjunction'}:
            raise TypeError('Only Conjunction and Disjunction are valid subclasses.')

    def __post_init__(self):
        if not self.predicates:
            raise ValueError(f'Cannot create empty {type(self).__name__}.')

    def referenced_identifiers(self) -> set[Identifier]:
        result = set()
        for predicate in self.predicates:
            result.update(predicate.referenced_identifiers())
        return result

    def _construct_instance(self, predicates: Iterable[Predicate]) -> Self:
        return self.__class__(tuple(predicates))

    def raise_from(self, on: Identifier) -> Self:
        predicates = tuple(p.raise_from(on) for p in self.predicates)
        return self._construct_instance(predicates)

    def lower_onto(self, on: Identifier) -> Self:
        predicates = tuple(p.lower_onto(on) for p in self.predicates)
        return self._construct_instance(predicates)

    def normalize(self) -> Predicate:
        normalized_predicates: List[Predicate] = []
        for predicate in self.predicates:
            normalized_predicate = predicate.normalize()
            if isinstance(normalized_predicate, self.__class__):  # unfold nested.
                normalized_predicates.extend(normalized_predicate.predicates)
            else:
                normalized_predicates.append(normalized_predicate)

        if len(normalized_predicates) == 1:  # avoid unnecessary nesting.
            return normalized_predicates[0]
        return self._construct_instance(normalized_predicates)


@dataclass(frozen=True)
class Conjunction(GenericJunction):
    """A conjunction of Predicates. See GenericJunction for implementation."""


@dataclass(frozen=True)
class Disjunction(GenericJunction):
    """A disjunction of Predicates. See GenericJunction for implementation."""


@dataclass(frozen=True)
class Token:
    identifier: Identifier = field(default_factory=Identifier)
    attributes: Optional[Predicate] = None


@dataclass(frozen=True)
class Dependency:
    src: Identifier
    dst: Identifier

    def referenced_identifiers(self) -> set[Identifier]:
        return {self.src, self.dst}


class Constraint(ABC):
    Order: ClassVar[type['Constraint']]
    Distance: ClassVar[type['Distance']]

    def referenced_identifiers(self) -> set[Identifier]:
        return set(self)

    @abstractmethod
    def __iter__(self): ...

    @staticmethod
    def order(a: Identifier, b: Identifier) -> 'Constraint':
        """
        Creates a new constraint stating that the first token must come
        before the second token in arrangements.
        :param a: Identifier of the first token.
        :param b: Identifier of the second token.
        :return: A new constraint.
        """
        return OrderConstraint(a, b)

    @staticmethod
    def distance(a: Identifier, b: Identifier):
        """
        Returns an object to encode distance constraints between two tokens.
        Distance constraints are created using comparison operators on the returned object.

        # Creates a constraint that the distance between a and b must be less than 3.
        distance(a, b) < 3

        :param a: Identifier of the first token.
        :param b: Identifier of the second token.
        :return: A new object to create distance constraints.
        """

        def make_constraint(order, dist: int) -> 'Constraint':
            return DistanceConstraint(a, b, order, dist)

        class Dist:
            def __eq__(self, other: int) -> 'Constraint':
                return make_constraint(Compare.EQ, other)

            def __ne__(self, other: int) -> 'Constraint':
                return make_constraint(Compare.NE, other)

            def __lt__(self, other: int) -> 'Constraint':
                return make_constraint(Compare.LT, other)

            def __gt__(self, other: int) -> 'Constraint':
                return make_constraint(Compare.GT, other)

            def __le__(self, other: int) -> 'Constraint':
                return self < (other + 1)

            def __ge__(self, other: int) -> 'Constraint':
                return self > (other + 1)

        return Dist()


class Compare(StrEnum):
    EQ = '='
    NE = '#'
    LT = '<'
    GT = '>'


@dataclass(frozen=True)
class DistanceConstraint(Constraint):
    a: Identifier
    b: Identifier
    order: Compare
    distance: int

    @override
    def __iter__(self):
        yield self.a
        yield self.b


Constraint.Distance = DistanceConstraint


@dataclass(frozen=True)
class OrderConstraint(Constraint):
    fst: Identifier
    snd: Identifier

    @override
    def __iter__(self):
        yield self.fst
        yield self.snd


Constraint.Order = OrderConstraint


@dataclass(frozen=True)
class Query:
    """
    A combination of tokens, dependencies, predicates and constraints.
    Optionally assigned an explicit identifier.
    """

    tokens: Collection[Token] = field(default_factory=set)
    dependencies: Collection[Dependency] = field(default_factory=set)
    constraints: Collection[Constraint] = field(default_factory=set)
    predicates: Collection[Predicate] = field(default_factory=set)
    identifier: Identifier = field(default_factory=Identifier)

    def __post_init__(self):
        defined_identifiers: Set[Identifier] = set()
        # Unlike graph-matching systems, we allow every identifier only once, as every identifier
        # is attached to a token and every token is unique.
        # The translation layer should handle unifying multiple references to the same identifier.

        for token in self.tokens:
            # Don't report identifiers here, since they are synthetic and meaningless for users.
            assert (
                token.identifier not in defined_identifiers
            ), 'Multiple tokens share the same identifier.'
            defined_identifiers.add(token.identifier)

        # Collect all identifiers referenced in query.
        referenced_identifiers = flatmap_set(self.constraints, lambda c: c.referenced_identifiers())
        referenced_identifiers |= flatmap_set(
            self.dependencies,
            lambda r: r.referenced_identifiers(),
        )
        referenced_identifiers |= flatmap_set(
            self.predicates,
            lambda p: p.referenced_identifiers(),
        )
        referenced_identifiers |= flatmap_set(
            self.tokens,
            lambda t: t.attributes.referenced_identifiers() if t.attributes else set(),
        )
        assert not (
            referenced_identifiers - defined_identifiers
        ), 'Query uses identifiers not defined by tokens.'


class SetOperator(StrEnum):
    CONJUNCTION = '&'
    DISJUNCTION = '|'
    SUBTRACTION = '-'


@dataclass(frozen=True)
class Operation:
    """A set operation performed on the results of two queries (or other operations)."""

    lhs: Identifier
    operator: SetOperator
    rhs: Identifier
    identifier: Identifier = field(default_factory=Identifier)


@dataclass(frozen=True)
class Recipe:
    """
    The full recipe to execute and combine a combination of Queries.
    The goal parameter determines what the recipe actually computes.

    Use the static Recipe.of_query(Query) method to construct a recipe of only one query.

    Alternatively, use the Recipe.Builder() class to construct a recipe step-by-step.
    """

    queries: Collection[Query]
    operations: Collection[Operation]
    goal: Identifier

    def __post_init__(self):
        identifiers: Set[Identifier] = set()
        for collection in [self.queries, self.operations]:
            for step in collection:
                assert (
                    step.identifier not in identifiers
                ), 'Multiple steps in recipe share the same identifier.'
                identifiers.add(step.identifier)

        for operation in self.operations:
            assert operation.lhs in identifiers, 'Step in recipe uses undefined query identifier.'
            assert operation.rhs in identifiers, 'Step in recipe uses undefined query identifier.'

    @staticmethod
    def of_query(query: Query) -> 'Recipe':
        return Recipe([query], [], query.identifier)

    def simple_representation(self) -> Optional[Query]:
        simple, *more = self.queries
        if not more and not self.operations:
            return simple
        return None

    def has_simple_representation(self) -> bool:
        return self.simple_representation() is not None

    def identifiers(self) -> Collection[Identifier]:
        query_identifiers = [query.identifier for query in self.queries]
        operation_identifiers = [operation.identifier for operation in self.operations]
        return operation_identifiers + query_identifiers

    def as_dict(self) -> dict[Identifier, Operation | Query]:
        result = {}
        for query in self.queries:
            result[query.identifier] = query
        for operation in self.operations:
            result[operation.identifier] = operation
        return result

    class Builder:
        def __init__(self):
            self.queries = list[Query]()
            self.operations = list[Operation]()
            self.explicit_goal: Optional[Identifier] = None

        def add_query(self, query: Query) -> Identifier:
            self.queries.append(query)
            return query.identifier

        def add_operation(self, lhs: Identifier, op: SetOperator, rhs: Identifier) -> Identifier:
            operation = Operation(lhs, op, rhs)
            self.operations.append(operation)
            return operation.identifier

        def set_goal(self, identifier: Identifier):
            self.explicit_goal = identifier

        def build(self) -> 'Recipe':
            if not self.explicit_goal:
                if len(self.queries) == 1:
                    goal = self.queries[0].identifier
                else:
                    goal = self.operations[-1].identifier
            else:
                goal = self.explicit_goal

            return Recipe(self.queries, self.operations, goal)



from collections import defaultdict
from dataclasses import dataclass, field
from typing import List, Self, Type, override

from antlr4 import CommonTokenStream, InputStream, TerminalNode
from antlr4.error.ErrorListener import ErrorListener

import cqp_tree.translation as ct
from cqp_tree.frontends.grew.antlr import GrewLexer, GrewParser


@dataclass
class ParseErrorListener(ErrorListener):
    errors: List[ct.InputError] = field(default_factory=list)

    @override
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        self.errors.append(ct.InputError(f'{line}, {column}', msg))


def parse(query: str) -> GrewParser.RequestContext:
    listener = ParseErrorListener()

    lexer = GrewLexer(InputStream(query))
    stream = CommonTokenStream(lexer)
    parser = GrewParser(stream)

    for antlr in [lexer, parser]:
        antlr.removeErrorListeners()
        antlr.addErrorListener(listener)

    result = parser.request()
    if listener.errors:
        raise ct.ParsingFailed(*listener.errors)

    return result


Environment = defaultdict[str, ct.Token]


def new_environment() -> Environment:
    return defaultdict(ct.Token)


@ct.translator('grew')
def translate_grew(grew: str) -> ct.Recipe:
    grew_request = parse(grew)
    return QueryBuilder().build(grew_request)


class QueryBuilder:

    def __init__(self, inherited: 'QueryBuilder' = None):
        self.dependencies = list[ct.Dependency]()
        self.constraints = list[ct.Constraint]()
        self.predicates = list[ct.Predicate]()

        self.environment = new_environment()

        if inherited:
            self.dependencies += inherited.dependencies
            self.constraints += inherited.constraints
            self.predicates += inherited.predicates

            for key, value in inherited.environment.items():
                # We never put attributes on tokens, so we are safe to copy references here.
                self.environment[key] = value

    def is_empty(self):
        return not self.environment

    def build_query(self) -> ct.Query:
        return ct.Query(
            tokens=self.environment.values(),
            dependencies=self.dependencies,
            constraints=self.constraints,
            predicates=self.predicates,
        )

    @staticmethod
    def query_operator(
        item: GrewParser.WithItemContext | GrewParser.WithoutItemContext,
    ) -> ct.SetOperator:
        return {
            GrewParser.WithItemContext: ct.SetOperator.CONJUNCTION,
            GrewParser.WithoutItemContext: ct.SetOperator.SUBTRACTION,
        }[type(item)]

    @staticmethod
    def build(request: GrewParser.RequestContext) -> ct.Recipe:
        plan = ct.Recipe.Builder()

        pattern = request.pattern().body()
        root_builder = QueryBuilder().translate_clauses(pattern)
        if root_builder.is_empty():
            # If pattern is empty, match an arbitrary token.
            query = ct.Query(tokens=[ct.Token()])
        else:
            query = root_builder.build_query()

        current = plan.add_query(query)
        goal = current

        for item in request.requestItem():
            builder = QueryBuilder(root_builder).translate_clauses(item.body())
            current = plan.add_query(builder.build_query())

            goal = plan.add_operation(goal, QueryBuilder.query_operator(item), current)

        plan.set_goal(goal)
        return plan.build()

    def translate_clauses(self, clauses: GrewParser.BodyContext) -> Self:
        for clause in clauses.clause():
            self.translate_clause(clause)
        return self

    @staticmethod
    def string_of_token(token: TerminalNode) -> str:
        return token.symbol.text

    @staticmethod
    def to_cqp_operator(grew: GrewParser.CompareContext) -> str:
        if isinstance(grew, GrewParser.EqualityContext):
            return '='

        if isinstance(grew, GrewParser.InequalityContext):
            return '!='

        assert False, f'Unknown operator: {type(grew)}'

    @staticmethod
    def wrap(
        predicates: List[ct.Predicate], ctor: Type[ct.Conjunction | ct.Disjunction]
    ) -> ct.Predicate:
        if len(predicates) > 1:
            return ctor(predicates)
        return predicates[0]

    def translate_clause(self, clause: GrewParser.ClauseContext):
        if isinstance(clause, GrewParser.NodeClauseContext):
            token = self.environment[clause.label.text]
            features = [
                translated_predicate.raise_from(token.identifier)
                for fs in clause.featureStructure()
                if (translated_predicate := self.to_predicate(fs))
            ]

            # Only add predicate if features are present.
            if features:
                predicate = self.wrap(features, ct.Disjunction)
                predicate.raise_from(token.identifier)
                self.predicates.append(predicate)

        elif isinstance(clause, GrewParser.EdgeClauseContext):
            src = self.environment[clause.src.text].identifier
            dst = self.environment[clause.dst.text].identifier

            dependency = ct.Dependency(src, dst)
            self.dependencies.append(dependency)

            arrow = clause.arrow()
            if isinstance(arrow, GrewParser.SimpleArrowContext):
                return  # No dependency types specified. Nothing else to do.

            deprel = ct.Attribute(dst, 'deprel')
            deptypes = [self.to_operand(dt) for dt in arrow.edgeTypes().literal()]
            if isinstance(arrow, GrewParser.PositiveArrowContext):
                dependency_constraint = self.wrap(
                    [ct.Comparison(deprel, '=', deptype) for deptype in deptypes],
                    ct.Disjunction,
                )

            elif isinstance(arrow, GrewParser.NegatedArrowContext):
                dependency_constraint = self.wrap(
                    [ct.Comparison(deprel, '!=', deptype) for deptype in deptypes],
                    ct.Conjunction,
                )

            else:
                raise ct.NotSupported(f'Unknown arrow type: {type(arrow)}')

            self.predicates.append(dependency_constraint)

        elif isinstance(clause, GrewParser.ConstraintClauseContext):
            lhs = self.to_operand(clause.lhs)
            rhs = self.to_operand(clause.rhs)
            predicate = ct.Comparison(
                lhs,
                self.to_cqp_operator(clause.compare()),
                rhs,
            )

            self.predicates.append(predicate)

        elif isinstance(clause, GrewParser.OrderClauseContext):
            lhs = self.environment[clause.lhs.text].identifier
            rhs = self.environment[clause.rhs.text].identifier

            self.constraints.append(ct.Constraint.order(lhs, rhs))
            if isinstance(clause.order(), GrewParser.ImmediatePrecedenceContext):
                self.constraints.append(ct.Constraint.distance(lhs, rhs) == 1)

    def to_predicate(
        self,
        grew: GrewParser.FeatureContext | GrewParser.FeatureStructureContext,
    ) -> ct.Predicate | None:
        # Checking for existence of feature: Tense
        if isinstance(grew, GrewParser.PresenceContext):
            attribute_name = self.string_of_token(grew.Identifier())
            return ct.Exists(ct.Attribute(None, attribute_name))

        # Checking for absence of feature: !Tense
        if isinstance(grew, GrewParser.AbsenceContext):
            attribute_name = self.string_of_token(grew.Identifier())
            return ct.Negation(ct.Exists(ct.Attribute(None, attribute_name)))

        # Requirement of feature to be certain value: Tense <> A|B|C
        if isinstance(grew, GrewParser.RequiresContext):
            attribute_name = self.string_of_token(grew.Identifier())
            attribute = ct.Attribute(None, attribute_name)

            alternatives = [self.to_operand(feature) for feature in grew.featureValue()]
            comparison = grew.compare()
            if isinstance(comparison, GrewParser.EqualityContext):
                return self.wrap(
                    [ct.Comparison(attribute, '=', alt) for alt in alternatives],
                    ct.Disjunction,
                )

            if isinstance(comparison, GrewParser.InequalityContext):
                return self.wrap(
                    [ct.Comparison(attribute, '!=', alt) for alt in alternatives],
                    ct.Conjunction,
                )

            raise ct.NotSupported(f'Unknown comparison type: {type(comparison)}')

        # Translate entire feature structure: [Tense, !Number, lemma = re"a.*"]
        if isinstance(grew, GrewParser.FeatureStructureContext):
            if not grew.feature():
                return None

            features = [self.to_predicate(feature) for feature in grew.feature()]
            return self.wrap(features, ct.Conjunction)

        assert False, f'Unknown predicate: {type(grew)}'

    def to_operand(
        self,
        grew: GrewParser.LiteralContext | GrewParser.FeatureValueContext,
    ) -> ct.Operand:
        # Proper string like: "aßσþ"
        if isinstance(grew, GrewParser.UnicodeStringContext):
            text = self.string_of_token(grew.String())
            return ct.Literal(text)

        # Dependency type with subtype: nsubj:pass
        if isinstance(grew, GrewParser.SubtypeContext):
            sup = self.string_of_token(grew.Identifier(0))
            sub = self.string_of_token(grew.Identifier(1))
            return ct.Literal(f'"{sup}:{sub}"')

        # Identifier used as simple string: Tense
        if isinstance(grew, GrewParser.SimpleStringContext):
            text = self.string_of_token(grew.Identifier())
            return ct.Literal(f'"{text}"')

        # Regular expression: re"a.*"
        if isinstance(grew, GrewParser.RegexContext):
            text = self.string_of_token(grew.String())
            return ct.Literal(text, represents_regex=True)

        # PCRE expression: /a.*/i
        if isinstance(grew, GrewParser.PCREContext):
            raise ct.NotSupported('PCRE expressions are not yet supported.')

        # Attribute of another token: X.upos
        if isinstance(grew, GrewParser.AttributeContext):
            instance = self.string_of_token(grew.Identifier(0))
            attribute_name = self.string_of_token(grew.Identifier(1))
            return ct.Attribute(self.environment[instance].identifier, attribute_name)

        # Literal used as part of feature structure, covered by other cases.
        if isinstance(grew, GrewParser.ValueContext):
            return self.to_operand(grew.literal())

        assert False, f'Unknown operand: {type(grew)}'



from dataclasses import dataclass
from typing import Any

from cqp_tree.utils import NonEmpty


@dataclass(frozen=True)
class InputError:
    """
    Error discovered while parsing input.
    """

    position: Any
    message: str

    def __repr__(self) -> str:
        return f'{self.position}: {self.message}'


class ParsingFailed(Exception):
    """
    Exception raised when parsing fails.
    """

    errors: NonEmpty[InputError]

    def __init__(self, *errors: InputError):
        assert errors, 'Expected at least 1 InputError as an argument.'
        super().__init__(f'Parsing failed. Detected {len(errors)} error(s).')
        self.errors = tuple(errors)


class NotSupported(Exception):
    """
    Exception raised when a query construct cannot (yet) be translated.
    """
